/* Searches using  Horspool's algorithm adapted to 
search inside a text compressed with End-Tagged Dense Code.
Lightweight Natural Language Text Compression: Information Retrieval 2006

Programmed by Antonio Faria.

Author's contact: Antonio Faria, Databases Lab, University of
A Corua. Campus de Elvia s/n. Spain  fari@udc.es

Copyright (C) 2006  Nieves R. Brisaboa, Gonzalo Navarro, Antonio Faria and Jos R. Param
Author's contact: antonio.fari@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


/*-----------------------------------------------------------------------
  search.c:  Implements the 2 following functions:
  			- searchInBuffer (searching for a pattern in a buffer)
  			- searchFullFile (leading the process of searching in all the buffers of a file)
 ------------------------------------------------------------------------*/
 
#include "searcher.h"


/*------------------------------------------------------------------
   Given a codeword stored as a number, it 
   obtains the number of bytes and writes they to *code as an array.	
------------------------------------------------------------------*/
void getBytes(unsigned char *code, unsigned long number, int numBytes) {
	unsigned long tmp;
	int i;
	char *pt;
	
	tmp = LE_LONG32(number);
	tmp = LINE_UP (tmp,numBytes);
	pt = (unsigned char *)&tmp;
	
	for (i=0;i<numBytes;i++) {
		code[i]= pt[i];				
	}	
}

/*------------------------------------------------------------------
 GetCode
 gets the codeword of a word ranked "position"
 returns an unsigned long value (up to 4 bytes codewords) in "*codeword"
         the number of bytes of the codeword.
------------------------------------------------------------------*/
void getCode(register unsigned long position,
						register unsigned long *codeword, register unsigned long *k){
	unsigned long despl;

	*k=1;
	*codeword = (position& 0x7F);
	position  >>= 7;
	despl=8;

	while (position > 0) {
		position -=1;
		*codeword |= (( (position & 0x7F) | 0x80 )<<despl );
		despl += 8;
		position  >>= 7;
		(*k) ++;
	}
}

/*------------------------------------------------------------------
 Prints a code: debuging purposes only
------------------------------------------------------------------*/
void printCodigo (byte *cw, int size) {
	int i;
	byte *c;
	c=cw;
	fprintf(stderr,"\nINICIO PRINTCODIGO codeSize %d: CODIGO:",size);
	for (i=0;i<size;i++) {fprintf(stderr,"[%d]",*(c++));}
	fflush (stderr);
}


/*------------------------------------------------------------------
  Initializes Horspool's d-vector (jumps for each letter)
 ------------------------------------------------------------------*/

void inicialize_d(byte *code, unsigned int len, byte *d) {
	unsigned int c,k;

	for (c=0;c<256;c++) {d[c]=len;}

	for (k=0; k<len-1; k++) {
		c=code[k];
		d[c] = MIN ( d[c], len - k-1);
	}
}



/*------------------------------------------------------------------
 Performs search process inside a Buffer of text.
 ------------------------------------------------------------------ */
unsigned int searchInBuffer (BufferRead B, byte **begin, byte **end,unsigned long pos,
					 byte *code, unsigned int len, byte *d ){

	int matches = 0;
	register byte *cend;
	register byte *cpos;
	register byte *cj;
	register int j;
	register unsigned int m;
		
	m=len;
	
	cpos=*begin + pos;
	cend=*end;

	while ( 1 )   {
		j=m-1;
		
		/*** skip loop  ***/		
		while ( *cpos != code[j] )  {cpos += d[*cpos];}
			
		/*** normal loop  ***/
		cj= cpos-1;
		j--;
				
		while ((j>=0) && ((*cj) == code[j])) {	j--;cj--;}  //check backwards
		
		if (j==-1) {
			if (cpos >=cend) {
				cj=cpos -1;
				while (*(cj) >= 128) {  //move cpos until previous stopper
					cj--;
						}
				cpos = cj; 
				cpos += d[*cj]; //calculates next jump before exitting						
				*begin = cpos - B->minDataInBuffer;	
				return matches;					
			}
						
			//a true match ?			
			if ((*cj) <128) {  //if the prev byte is a stopper 
				matches ++;			
			}			
		}
			
		cpos += d[*cpos];
			
	} //while ( 1 )

	*begin = cpos - B->minDataInBuffer;
	return matches;
}


/*------------------------------------------------------------------
 * Performs search process over the whole file.
------------------------------------------------------------------ */
unsigned int searchFullFile (char *infile, byte *code, int len){
	BufferRead B;
	byte d[256];
	int f;
	int matches=0;
	byte AUX[4];  	//Used to save the 1-4 bytes that are modified in the buffer 
						//to put a break code at the end of the buffer.

	byte *text,*top, *nextbuf;  //pointers to the input buffer

	//initializes Horspool jumps vector "d[]" (used to jump forwards through the file)
	inicialize_d(code, len, d);

	// Initialization of the structure Buffer used to read the file
	f = open (infile, O_RDONLY);
	if (f == -1) {
		fprintf(stderr,"Cannot read file %s",infile);
		exit(0);
	}

	B= bufCreate(len+1);
	bufSetFile (B,f);

	// Search process. ................

	fprintf(stderr, "\n STARTING SEARCH......");
	{ unsigned long i;
		byte *NEWEND;

		byte *pbeg,*pend,*codeStart;
		unsigned long pos;

		pos=len-1;

		//process the whole file by reading one buffer of data at a time and then performing
		//set-horspool (modified) over those buffers of data.
		while (!bufEmpty(B) ) {

			// set the area of the buffer that will be processed.
			bufCurrent (B,(byte**)&text,(byte**)&top);
			pbeg = text; 
			pend = top;

			// copies a searched pattern (the first one) to the end of the buffer as a BREAK-CODE
			// in order to avoid the need of  checking for the "end of buffer" in the search process.
			// Since some bytes at the end of the buffer are replaced, the are saved in AUX first.
			NEWEND = pend+1;
			for (i=0;i<len;i++){ 
				AUX[i]=NEWEND[i];
				NEWEND[i]=code[i]; 
			}
					
			// Searching inside the buffer 
			matches += searchInBuffer (B, &pbeg, &pend,pos, code,len, d );

			//the bytes at the end of the buffer are restored to the buffer from AUX.
			NEWEND = pend+1;
			for (i=0;i<len;i++){ 
				NEWEND[i]=AUX[i];					
			}
	
			// Reads the next buffer of data that will be processed in the next iteration
			pos = B->dsize > B->minDataInBuffer? B->minDataInBuffer: B->dsize;
			if (bufEof(B)) nextbuf=top;
			else nextbuf=pbeg;

			bufLoad (B, (char*)nextbuf);

		}//while
	}
	fprintf(stderr, "... search FINISHED with %d matches!!\n", matches);;fflush(stderr);
		
	//Frees memory...........
	bufDestroy(B);
	return matches;
} //searchFullFile




































/************************************************************************************************/
/*******   OLD VERSIONS ............    ***************************************************/
/************************************************************************************************/


/************************************************************************************************/
/*------------------------------------------------------------------
 Performs search process inside a Buffer of text.
 ------------------------------------------------------------------ */
unsigned int searchInBuffer_old (BufferRead B, byte **begin, byte **end,unsigned long pos,
					 byte *code, unsigned int len, byte *d ){


	int matches = 0;
	register byte *cend;
	register byte *cpos;
	register byte *cj;
	register int j;
	register unsigned int m;
		
	m=len;
	
	cpos=*begin + pos;
	cend=*end;
	register unsigned int s=128;

	while ( 1 )   {
		j=m-1;
		
		/*** skip loop  ***/		
		while ( *cpos != code[j] ) {
				#ifdef CONTADOR_BYTES_ENABLED    
					LEIDOS += d[*cpos];			
				#endif
			cpos += d[*cpos];
		}
			
		/*** normal loop  ***/
		cj= cpos-1;
		j--;
		//cj=cpos;
				
		while ((j>=0) && ((*cj) == code[j])) {
			j--;cj--;
		}
		
		if (j==-1) {
			if (cpos >=cend) {
				cj=cpos -1;
						#ifdef CONTADOR_BYTES_ENABLED
					LEIDOS --;
						#endif
				while (*(cj) >= s) {  //move cpos until previous stopper
					cj--;
						#ifdef CONTADOR_BYTES_ENABLED
					LEIDOS --;
						#endif
						}
				cpos = cj+1; 
						#ifdef CONTADOR_BYTES_ENABLED
				LEIDOS++;
						#endif
										
				*begin = cpos - B->minDataInBuffer;	
				//fprintf(stderr,"\n salto horspool de 1 buffer: cpos %ld ,cend %ld",cpos,cend);fflush(stderr);							 
				return matches;					
			}
			
//			if (cj < *begin) fprintf(stderr,"\n AAAAAIIIII ");
			
			//a true match ?
			
			//cj--;
			if ((*cj) <s) {
				matches ++;			
			}			
		}
			
			#ifdef CONTADOR_BYTES_ENABLED
		LEIDOS += d[*cpos];	
			#endif
		cpos += d[*cpos];
			
	} //while ( 1 )

	*begin = cpos - B->minDataInBuffer;
	return matches;
}


/*------------------------------------------------------------------
 * Performs search process over the whole file.
------------------------------------------------------------------ */
unsigned int searchFullFile_old (char *infile, byte *code, int len){
	BufferRead B;
	byte d[256];
	int f;
	int matches=0;
	byte AUX[4];  	//Used to save the 1-4 bytes that are modified in the buffer 
					//to put a break code at the end of the buffer.

	byte *text,*top, *nextbuf;  //pointers to the input buffer


		#ifdef CONTADOR_BYTES_ENABLED
	LEIDOS = 0;
		#endif


	//initializes Horspool jumps vector "d[]" (used to jump forwards through the file)
	inicialize_d(code, len, d);


	// Initialization of the structure Buffer used to read the file
	f = open (infile, O_RDONLY);
	if (f == -1) {
		fprintf(stderr,"Cannot read file %s",infile);
		exit(0);
	}

	B= bufCreate(len+1);
	bufSetFile (B,f);

	// Search process. ................

	fprintf(stderr, "\n STARTING SEARCH......");
	{ unsigned long i;
		byte *NEWEND;

		byte *pbeg,*pend,*codeStart;
		unsigned long pos;

		pos=len-1;

			#ifdef CONTADOR_BYTES_ENABLED
		LEIDOS = pos;
			#endif

		//process the whole file by reading one buffer of data at a time and then performing
		//set-horspool (modified) over those buffers of data.
		while (!bufEmpty(B) ) {

			// set the area of the buffer that will be processed.
			bufCurrent (B,(byte**)&text,(byte**)&top);
			pbeg = text; 
			pend = top;

			// copies a searched pattern (the first one) to the end of the buffer as a BREAK-CODE
			// in order to avoid the need of  checking for the "end of buffer" in the search process.
			// Since some bytes at the end of the buffer are replaced, the are saved in AUX first.
			NEWEND = pend+1;
			for (i=0;i<len;i++){ 
				AUX[i]=NEWEND[i];
				NEWEND[i]=code[i]; 
			}
					
			// Searching inside the buffer 
			matches += searchInBuffer (B, &pbeg, &pend,pos, code,len, d );

			//the bytes at the end of the buffer are restored to the buffer from AUX.
			NEWEND = pend+1;
			for (i=0;i<len;i++){ 
				NEWEND[i]=AUX[i];					
			}

			
			// Reads the next buffer of data that will be processed in the next iteration
			pos = B->dsize > B->minDataInBuffer? B->minDataInBuffer: B->dsize;
			if (bufEof(B)) nextbuf=top;
			else nextbuf=pbeg;

			//fprintf(stderr, "\n...pasaba por aqu1 nextbuf %ld top %ld pbeg %ld\n",nextbuf,top,pbeg);;fflush(stderr);
			bufLoad (B, (char*)nextbuf);
			//fprintf(stderr, "\n...pasaba por aqu2 \n");;fflush(stderr);

			#ifdef CONTADOR_BYTES_ENABLED    
				fprintf(stderr,"\n TOTAL BYTES READ (or jumped): %ld \n",	LEIDOS);;fflush(stderr);
			#endif			
		}//while
	}

	fprintf(stderr, "... search FINISHED with %d matches!!\n", matches);;fflush(stderr);

	// Summarizes the results obtained.
	// If login is enabled, it writes, for each pattern, the number of occurrencias found.
		
	#ifdef LOGGING_ENABLED     
		FILE *fLOG2;
	
		if ((fLOG2 = fopen("log2","a")) == NULL)
			{
			  fprintf(stderr,
				  "Error en la apertura del fichero de tiempos %s\n",
				  "tiempos");
			  exit(0);
			}

			fprintf(fLOG2,"%5d.- %6d\n",matches);
		
			
		fflush(fLOG2);
		fclose(fLOG2);
	#endif 	

	#ifdef CONTADOR_BYTES_ENABLED    
		fprintf(stderr,"\n TOTAL BYTES READ (or jumped): %ld \n",	LEIDOS);
	#endif

	//Frees memory...........
	bufDestroy(B);
	return matches;
} //searchFullFile
